@@ -1,4 +1,5 @@
-/* Copyright (c) 2021-2024 The Khronos Group Inc.
+/* Copyright (c) 2021-2025 The Khronos Group Inc.
+ * Copyright (c) 2025 Arm Limited.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -17,16 +18,19 @@
 
 #pragma once
 
+#include <vulkan/vulkan_core.h>
 #include <cassert>
 #include <cstdlib>
 #include <cstring>
+#include <string>
 #include <vector>
 
+#include "containers/custom_containers.h"
 #include "state_tracker/shader_instruction.h"
 #include "state_tracker/state_object.h"
 #include "state_tracker/sampler_state.h"
 #include <spirv/unified1/spirv.hpp>
-#include "spirv-tools/optimizer.hpp"
+#include "containers/limits.h"
 
 namespace vvl {
 class Pipeline;
@@ -38,9 +42,19 @@ struct Module;
 
 static constexpr uint32_t kInvalidValue = std::numeric_limits<uint32_t>::max();
 
-// Need to find a way to know if actually array lenght of zero, or a runtime array.
+// Need to find a way to know if actually array length of zero, or a runtime array.
 static constexpr uint32_t kRuntimeArray = std::numeric_limits<uint32_t>::max();
 
+struct LocalSize {
+    uint32_t x = 0;
+    uint32_t y = 0;
+    uint32_t z = 0;
+
+    std::string ToString() const {
+        return "x = " + std::to_string(x) + ", y = " + std::to_string(y) + ", z = " + std::to_string(z);
+    }
+};
+
 // This is the common info for both OpDecorate and OpMemberDecorate
 // Used to keep track of all decorations applied to any instruction
 struct DecorationBase {
@@ -97,66 +111,87 @@ struct DecorationSet : public DecorationBase {
 
 // Tracking of OpExecutionMode / OpExecutionModeId values
 struct ExecutionModeSet {
-    enum FlagBit {
-        output_points_bit = 1 << 0,
-        point_mode_bit = 1 << 1,
-        post_depth_coverage_bit = 1 << 2,
-        local_size_bit = 1 << 3,
-        local_size_id_bit = 1 << 4,
-        iso_lines_bit = 1 << 5,
-        xfb_bit = 1 << 6,
-        early_fragment_test_bit = 1 << 7,
-        subgroup_uniform_control_flow_bit = 1 << 8,
+    enum FlagBit : uint64_t {
+        output_points_bit = 1ull << 0,
+        output_lines_bit = 1ull << 1,
+        output_triangle_bit = 1ull << 2,
 
-        signed_zero_inf_nan_preserve_width_16 = 1 << 9,
-        signed_zero_inf_nan_preserve_width_32 = 1 << 10,
-        signed_zero_inf_nan_preserve_width_64 = 1 << 11,
-        denorm_preserve_width_16 = 1 << 12,
-        denorm_preserve_width_32 = 1 << 13,
-        denorm_preserve_width_64 = 1 << 14,
-        denorm_flush_to_zero_width_16 = 1 << 15,
-        denorm_flush_to_zero_width_32 = 1 << 16,
-        denorm_flush_to_zero_width_64 = 1 << 17,
-        rounding_mode_rte_width_16 = 1 << 18,
-        rounding_mode_rte_width_32 = 1 << 19,
-        rounding_mode_rte_width_64 = 1 << 20,
-        rounding_mode_rtz_width_16 = 1 << 21,
-        rounding_mode_rtz_width_32 = 1 << 22,
-        rounding_mode_rtz_width_64 = 1 << 23,
+        subdivision_iso_lines_bit = 1ull << 3,
+        subdivision_triangle_bit = 1ull << 4,
+        subdivision_quad_bit = 1ull << 5,
 
-        depth_replacing_bit = 1 << 24,
-        stencil_ref_replacing_bit = 1 << 25,
+        vertex_order_cw_bit = 1ull << 6,
+        vertex_order_ccw_bit = 1ull << 7,
 
-        fp_fast_math_default = 1 << 26,
+        spacing_equal_bit = 1ull << 8,
+        spacing_fractional_even_bit = 1ull << 9,
+        spacing_fractional_odd_bit = 1ull << 10,
+
+        point_mode_bit = 1ull << 11,
+        post_depth_coverage_bit = 1ull << 12,
+        local_size_bit = 1ull << 13,
+        local_size_id_bit = 1ull << 14,
+        xfb_bit = 1ull << 15,
+        early_fragment_test_bit = 1ull << 16,
+        subgroup_uniform_control_flow_bit = 1ull << 17,
+
+        signed_zero_inf_nan_preserve_width_16 = 1ull << 18,
+        signed_zero_inf_nan_preserve_width_32 = 1ull << 19,
+        signed_zero_inf_nan_preserve_width_64 = 1ull << 20,
+        denorm_preserve_width_16 = 1ull << 21,
+        denorm_preserve_width_32 = 1ull << 22,
+        denorm_preserve_width_64 = 1ull << 23,
+        denorm_flush_to_zero_width_16 = 1ull << 24,
+        denorm_flush_to_zero_width_32 = 1ull << 25,
+        denorm_flush_to_zero_width_64 = 1ull << 26,
+        rounding_mode_rte_width_16 = 1ull << 27,
+        rounding_mode_rte_width_32 = 1ull << 28,
+        rounding_mode_rte_width_64 = 1ull << 29,
+        rounding_mode_rtz_width_16 = 1ull << 30,
+        rounding_mode_rtz_width_32 = 1ull << 31,
+        rounding_mode_rtz_width_64 = 1ull << 32,
+
+        depth_replacing_bit = 1ull << 33,
+        stencil_ref_replacing_bit = 1ull << 34,
+
+        fp_fast_math_default = 1ull << 35,
+
+        derivative_group_linear = 1ull << 36,
+        derivative_group_quads = 1ull << 37,
+
+        geometry_input_points_bit = 1ull << 38,
+        geometry_input_line_bit = 1ull << 39,
+        geometry_input_line_adjacency_bit = 1ull << 40,
+        geometry_input_triangle_bit = 1ull << 41,
+        geometry_input_triangle_adjacency_bit = 1ull << 42,
     };
 
     // bits to know if things have been set or not by a Decoration
-    uint32_t flags = 0;
-
-    VkPrimitiveTopology input_primitive_topology = VK_PRIMITIVE_TOPOLOGY_MAX_ENUM;
-    VkPrimitiveTopology primitive_topology = VK_PRIMITIVE_TOPOLOGY_MAX_ENUM;
+    uint64_t flags = 0;
 
     // SPIR-V spec says only LocalSize or LocalSizeId can be used, so can share
-    uint32_t local_size_x = kInvalidValue;
-    uint32_t local_size_y = kInvalidValue;
-    uint32_t local_size_z = kInvalidValue;
+    LocalSize local_size = {kInvalidValue, kInvalidValue, kInvalidValue};
 
-    uint32_t output_vertices = vvl::kU32Max;
+    uint32_t output_vertices = kInvalidValue;
     uint32_t output_primitives = 0;
     uint32_t invocations = 0;
 
-    uint32_t tessellation_subdivision = 0;
-    uint32_t tessellation_orientation = 0;
-    uint32_t tessellation_spacing = 0;
-
     void Add(const Instruction &insn);
     bool Has(FlagBit flag_bit) const { return (flags & flag_bit) != 0; }
+
+    // Helpers for the various input/output stuff Geom/Tess/Mesh has
+    uint32_t GetTessellationSubdivision() const;
+    uint32_t GetTessellationOrientation() const;
+    uint32_t GetTessellationSpacing() const;
+    VkPrimitiveTopology GetTessellationEvalOutputTopology() const;
+    VkPrimitiveTopology GetGeometryInputTopology() const;
+    VkPrimitiveTopology GetGeometryMeshOutputTopology() const;
 };
 
 struct AtomicInstructionInfo {
     uint32_t storage_class;
     uint32_t bit_width;
-    uint32_t type;  // ex. OpTypeInt
+    uint32_t type;             // ex. OpTypeInt
     uint32_t vector_size = 0;  // 0 for scalar, otherwise number of components
 };
 
@@ -219,34 +254,31 @@ using FuncParameterMap = vvl::unordered_map<uint32_t, std::vector<uint32_t>>;
 
 // Represents the OpImage* instructions and how it maps to the variable
 // This is created in the Module but then used with VariableBase objects
-struct ImageAccess {
-    const Instruction &image_insn;  // OpImage*
+//
+// This is "static" because it is useless for descriptor indexing.
+// If 2 different OpImage* access the same variable (that is an array of descriptors), both must be valid, but for descriptor
+// indexing, these attributes aren't tied to the whole variable, just to the index accessed
+struct StaticImageAccess {
     std::vector<const Instruction *> variable_image_insn;
     // If there is a OpSampledImage there will also be a sampler variable
     std::vector<const Instruction *> variable_sampler_insn;
     // incase uncaught set of SPIR-V instruction is found, skips validating instead of crashing
     bool valid_access = true;
 
-    bool is_dref = false;
-    bool is_sampler_implicitLod_dref_proj = false;
-    bool is_sampler_sampled = false;  // OpImageSample* or OpImageSparseSample*
-    bool is_not_sampler_sampled = false;
-    bool is_sampler_bias_offset = false;
-    bool is_sampler_offset = false;  // ConstOffset or Offset (not ConstOffsets)
-    bool is_sign_extended = false;
-    bool is_zero_extended = false;
+    const ImageInstruction image_insn;
+
     uint32_t access_mask = AccessBit::empty;
 
     uint32_t image_access_chain_index = kInvalidValue;    // OpAccessChain's Index 0
     uint32_t sampler_access_chain_index = kInvalidValue;  // OpAccessChain's Index 0
     uint32_t texel_component_count = kInvalidValue;
 
-    ImageAccess(const Module &module_state, const Instruction &image_insn, const FuncParameterMap &func_parameter_map);
+    StaticImageAccess(const Module &module_state, const Instruction &insn, const FuncParameterMap &func_parameter_map);
 };
 
-// <Image OpVariable Result ID, [ImageAccess, ImageAccess, etc] > - used for faster lookup
-// Many ImageAccess can point to a single Image Variable
-using ImageAccessMap = vvl::unordered_map<uint32_t, std::vector<std::shared_ptr<const ImageAccess>>>;
+// <Image OpVariable Result ID, [StaticImageAccess, StaticImageAccess, etc] > - used for faster lookup
+// Many StaticImageAccess can point to a single Image Variable
+using StaticImageAccessMap = vvl::unordered_map<uint32_t, std::vector<std::shared_ptr<const StaticImageAccess>>>;
 // < Variable ID, [ OpAccessChain ] >
 // Allows for grouping the access chains by which variables they are actually accessing
 using AccessChainVariableMap = vvl::unordered_map<uint32_t, std::vector<const Instruction *>>;
@@ -294,12 +326,13 @@ enum NumericType {
     NumericTypeUint = 4,
 };
 uint32_t GetFormatType(VkFormat format);
-char const *string_NumericType(uint32_t type);
+const char *string_NumericType(uint32_t type);
 
 // Common info needed for all OpVariable
 struct VariableBase {
     const uint32_t id;
     const uint32_t type_id;
+    const uint32_t data_type_id;
     const spv::StorageClass storage_class;
     const DecorationSet &decorations;
     std::shared_ptr<const TypeStructInfo> type_struct_info;  // null if no struct type
@@ -334,6 +367,10 @@ struct VariableBase {
     // memory was read
     bool IsImageReadFrom() const { return access_mask & AccessBit::image_read; }
     bool IsImageWrittenTo() const { return access_mask & AccessBit::image_write; }
+    // Something like textureSize() will access the OpVariable, but not the image itself
+    bool IsImageAccessed() const { return access_mask & AccessBit::image_mask; }
+
+    bool IsUntyped() const { return data_type_id != 0; }
 
   private:
     static const char *FindDebugName(const VariableBase &variable, const DebugNameMap &debug_name_map);
@@ -357,7 +394,6 @@ struct StageInterfaceVariable : public VariableBase {
     const Instruction &base_type;
     const bool is_builtin;
     bool nested_struct;
-    bool physical_storage_buffer;
 
     const std::vector<InterfaceSlot> interface_slots;  // Only for User Defined variables
     const std::vector<uint32_t> builtin_block;
@@ -387,10 +423,13 @@ struct ResourceInterfaceVariable : public VariableBase {
     // Will be kRuntimeArray (non-zero) for runtime arrays
     uint32_t array_length;
 
-    bool is_sampled_image;  // OpTypeSampledImage
+    // OpTypeSampledImage (used for combined image samplers)
+    bool is_type_sampled_image;
 
-    // List of samplers that sample a given image. The index of array is index of image.
+    // The index of vector is index of image. (TODO - this doesn't work for GPU-AV)
     std::vector<vvl::unordered_set<SamplerUsedByImage>> samplers_used_by_image;
+    // workaround for YCbCr to track sampler variables until |samplers_used_by_image| is fixed
+    vvl::unordered_set<uint32_t> sampled_image_sampler_variable_ids;
 
     // For storage images - list of Texel component length the OpImageWrite
     std::vector<uint32_t> write_without_formats_component_count_list;
@@ -404,64 +443,60 @@ struct ResourceInterfaceVariable : public VariableBase {
     // most likly will be OpTypeImage, OpTypeStruct, OpTypeSampler, or OpTypeAccelerationStructureKHR
     const Instruction &base_type;
 
-    // This is true if either:
-    // - The descriptor is made up of an OpRuntimeArray
-    // - Not all OpAccessChains pointing to this descriptor or easily determined constant
-    bool is_dynamic_accessed;
-
-    // Sampled Type width of the OpTypeImage the variable points to, 0 if doesn't use the image
-    const uint32_t image_sampled_type_width;
+    // True if the Resource variable itself is runtime descriptor array
+    // Online example to showcase various arrays we do/don't care about here https://godbolt.org/z/h9jhsKaPn
+    bool is_runtime_descriptor_array;
 
     // All info regarding what will be validated from requirements imposed by the pipeline on a descriptor. These
     // can't be checked at pipeline creation time as they depend on the Image or ImageView bound.
     // That is perf-critical code and hashing if 2 variables have same info provides a 20% perf bonus
+    //
+    // This info can be broken down into two further parts:
+    //  1. Things that are tied to the OpVariable (ex, Image Dim can be found from the OpTypeImage)
+    //  2. Things that depend on the image access operation (ex, Dref is not known looking at a OpVariable)
+    // For Descriptor Indexing, we have to be careful about the 2nd case as it is possible to an OpVariable to contain descriptors
+    // in the array that match correctly depending on the index used. (For normal 1.0 descriptor, all access to the array must be
+    // valid for all indexes in the array)
     struct Info {
-        NumericType image_format_type;
+        // the 'format' operand of OpTypeImage as the corresponding Vulkan Format
+        VkFormat image_format{VK_FORMAT_UNDEFINED};
+        // the 'Sampled Type' operand of OpTypeImage,as a numeric type (float, uint, int)
+        NumericType image_sampled_type_numeric{NumericTypeUnknown};
+        // the 'Sampled Type' operand of OpTypeImage as the bit width (64 is the largest bit width in SPIR-V)
+        uint8_t image_sampled_type_width{0};
+
         spv::Dim image_dim;
         bool is_image_array;
         bool is_multisampled;
 
-        bool is_sampler_sampled{false};  // OpImageSample* or OpImageSparseSample*
-        bool is_not_sampler_sampled{false};
-        bool is_sampler_implicitLod_dref_proj{false};
-        bool is_sampler_bias_offset{false};
-        bool is_sampler_offset{false};        // ConstOffset or Offset (not ConstOffsets)
+        ImageInstruction image_insn;
+
         bool is_read_without_format{false};   // For storage images
         bool is_write_without_format{false};  // For storage images
-        bool is_dref{false};
-
-        // vkspec.html#spirvenv-image-signedness describes how SignExtend/ZeroExtend can be used per-access to adjust the Signedness
-        // Only need to check if one access has explicit signedness, mixing should be caught in spirv-val
-        bool is_sign_extended{false};  // if at least one access has SignExtended
-        bool is_zero_extended{false};  // if at least one access has ZeroExtended
 
         // If a variable is used as a function arguement, but never actually used, it will be found in EntryPoint::accessible_ids so
         // we need to have a dedicated mark if it was accessed.
         // We use this for variable hashing, but the VariableBase has the helper functions to read this value.
         uint32_t access_mask{AccessBit::empty};
     } info;
+    // For non descriptor indexing usages, this hash allows use to skip re-validating because a different VkImageView bound will
+    // result in the same outcome
     uint64_t descriptor_hash = 0;
-    bool IsImage() const { return info.image_format_type != NumericTypeUnknown; }
-
-    // If dealing with an image array, only check the indexes accesses
-    vvl::unordered_set<uint32_t> image_access_chain_indexes;
+    bool IsImage() const { return base_type.Opcode() == spv::OpTypeImage; }
 
     // Type of resource type (vkspec.html#interfaces-resources-storage-class-correspondence)
     bool is_storage_image{false};
     bool is_storage_texel_buffer{false};
     const bool is_storage_buffer;
     bool is_input_attachment{false};
+    bool is_storage_tensor{false};
 
     ResourceInterfaceVariable(const Module &module_state, const EntryPoint &entrypoint, const Instruction &insn,
-                              const ImageAccessMap &image_access_map, const AccessChainVariableMap &access_chain_map,
+                              const StaticImageAccessMap &image_access_map, const AccessChainVariableMap &access_chain_map,
                               const VariableAccessMap &variable_access_map, const DebugNameMap &debug_name_map);
 
   protected:
     static const Instruction &FindBaseType(ResourceInterfaceVariable &variable, const Module &module_state);
-    static bool IsDynamicAccessed(ResourceInterfaceVariable &variable, const Module &module_state,
-                                  const AccessChainVariableMap &access_chain_map);
-    static uint32_t FindImageSampledTypeWidth(const Module &module_state, const Instruction &base_type);
-    static NumericType FindImageFormatType(const Module &module_state, const Instruction &base_type);
     static bool IsStorageBuffer(const ResourceInterfaceVariable &variable);
 };
 
@@ -482,11 +517,19 @@ struct PushConstantVariable : public VariableBase {
                          const VariableAccessMap &variable_access_map, const DebugNameMap &debug_name_map);
 };
 
+struct TaskPayloadVariable : public VariableBase {
+    uint32_t size;
+
+    TaskPayloadVariable(const Module &module_state, const Instruction &insn, VkShaderStageFlagBits stage,
+                        const VariableAccessMap &variable_access_map, const DebugNameMap &debug_name_map);
+};
+
 // Represents a single Entrypoint into a Shader Module
 struct EntryPoint {
     // "A module must not have two OpEntryPoint instructions with the same Execution Model and the same Name string."
     // There is no single unique item for a single entry point
     const Instruction &entrypoint_insn;  // OpEntryPoint instruction
+    const bool is_data_graph;
     // For things like MeshNV vs MeshEXT, we need the execution_model
     const spv::ExecutionModel execution_model;
     const VkShaderStageFlagBits stage;
@@ -503,6 +546,8 @@ struct EntryPoint {
 
     // only one Push Constant block is allowed per entry point
     std::shared_ptr<const PushConstantVariable> push_constant_variable;
+    // For both Task and Mesh entry point, there can be one TaskPayloadWorkgroupEXT variable
+    std::shared_ptr<const TaskPayloadVariable> task_payload_variable;
     const std::vector<ResourceInterfaceVariable> resource_interface_variables;
     const std::vector<StageInterfaceVariable> stage_interface_variables;
     // Easier to lookup without having to check for the is_builtin bool
@@ -511,6 +556,8 @@ struct EntryPoint {
     // "User-defined Variable Interface" - vkspec.html#interfaces-iointerfaces-user
     std::vector<const StageInterfaceVariable *> user_defined_interface_variables;
 
+    // Map for quick reserve lookup of variables from the OpVariable Result ID
+    vvl::unordered_map<uint32_t, const ResourceInterfaceVariable *> resource_interface_variable_map;
     // Lookup map from Interface slot to the variable in that spot
     // spirv-val guarantees no overlap so 2 variables won't have same slot
     vvl::unordered_map<InterfaceSlot, const StageInterfaceVariable *, InterfaceSlot::Hash> input_interface_slots;
@@ -533,17 +580,21 @@ struct EntryPoint {
     bool has_passthrough{false};
     bool has_alpha_to_coverage_variable{false};  // only for Fragment shaders
 
-    EntryPoint(const Module &module_state, const Instruction &entrypoint_insn, const ImageAccessMap &image_access_map,
+    bool has_physical_storage_buffer_interface{false};
+
+    EntryPoint(const Module &module_state, const Instruction &entrypoint_insn, const StaticImageAccessMap &image_access_map,
                const AccessChainVariableMap &access_chain_map, const VariableAccessMap &variable_access_map,
                const DebugNameMap &debug_name_map);
 
+    bool HasBuiltIn(spv::BuiltIn built_in) const;
+
   protected:
     static vvl::unordered_set<uint32_t> GetAccessibleIds(const Module &module_state, EntryPoint &entrypoint);
-    static std::vector<StageInterfaceVariable> GetStageInterfaceVariables(const Module &module_state, const EntryPoint &entrypoint,
+    static std::vector<StageInterfaceVariable> GetStageInterfaceVariables(const Module &module_state, EntryPoint &entrypoint,
                                                                           const VariableAccessMap &variable_access_map,
                                                                           const DebugNameMap &debug_name_map);
     static std::vector<ResourceInterfaceVariable> GetResourceInterfaceVariables(const Module &module_state, EntryPoint &entrypoint,
-                                                                                const ImageAccessMap &image_access_map,
+                                                                                const StaticImageAccessMap &image_access_map,
                                                                                 const AccessChainVariableMap &access_chain_map,
                                                                                 const VariableAccessMap &variable_access_map,
                                                                                 const DebugNameMap &debug_name_map);
@@ -551,13 +602,13 @@ struct EntryPoint {
                                  const AccessChainVariableMap &access_chain_map);
 };
 
-// Info to capture while parsing the SPIR-V, but will only be used by ValidateSpirvStateless and don't need to save after
+// Info to capture while parsing the SPIR-V, but will only be used by SpirvValidator::Validate and don't need to save after
 struct StatelessData {
     // Used if the Shader Module is being passed in VkPipelineShaderStageCreateInfo
     std::shared_ptr<spirv::Module> pipeline_pnext_module;
 
     // These instruction mapping were designed to quickly find the few instructions without having to loop the entire pass
-    // In theory, these could be removed checked during the 2nd pass in ValidateSpirvStateless()
+    // In theory, these could be removed checked during the 2nd pass in SpirvValidator::Validate
     // TODO - Get perf numbers if better to understand if these make sense here
     std::vector<const Instruction *> read_clock_inst;
     std::vector<const Instruction *> atomic_inst;
@@ -571,6 +622,7 @@ struct StatelessData {
     bool has_builtin_fully_covered{false};
     bool has_invocation_repack_instruction{false};
     bool has_group_decoration{false};
+    bool has_ext_inst_with_forward_refs{false};  // OpExtInstWithForwardRefsKHR
 };
 
 // Represents a SPIR-V Module
@@ -580,7 +632,7 @@ struct Module {
     // The goal of this struct is to move everything that is ready only into here
     struct StaticData {
         StaticData() = default;
-        StaticData(const Module &module_state, StatelessData *stateless_data = nullptr);
+        StaticData(const Module &module_state, bool parse, StatelessData *stateless_data);
         StaticData &operator=(StaticData &&) = default;
         StaticData(StaticData &&) = default;
 
@@ -604,8 +656,10 @@ struct Module {
         // Find all decoration instructions to prevent relooping module later - many checks need this info
         std::vector<const Instruction *> decoration_inst;
         std::vector<const Instruction *> member_decoration_inst;
-        // Find all variable instructions to prevent relookping module later
+        // Find all variable instructions to build faster LUT
         std::vector<const Instruction *> variable_inst;
+        // Both variables and instruction explicitly accessing untyped variables
+        std::vector<const Instruction *> explicit_memory_inst;
         // For shader tile image - OpDepthAttachmentReadEXT/OpStencilAttachmentReadEXT/OpColorAttachmentReadEXT
         bool has_shader_tile_image_depth_read{false};
         bool has_shader_tile_image_stencil_read{false};
@@ -618,6 +672,8 @@ struct Module {
         uint32_t builtin_workgroup_size_id = 0;
 
         std::vector<const Instruction *> cooperative_matrix_inst;
+        std::vector<const Instruction *> cooperative_vector_inst;
+        std::vector<const Instruction *> emit_mesh_tasks_inst;
 
         std::vector<spv::Capability> capability_list;
         // Code on the hot path can cache capabilities for fast access.
@@ -626,6 +682,11 @@ struct Module {
         bool has_specialization_constants{false};
         bool uses_interpolate_at_sample{false};
 
+        // Will check if there is source debug information
+        // Won't save any other info and will retrieve the debug info if requested in a VU error message
+        bool using_legacy_debug_info{false};
+        uint32_t shader_debug_info_set_id = 0;  // non-zero means shader has NonSemantic.Shader.DebugInfo.100
+
         // EntryPoint has pointer references inside it that need to be preserved
         std::vector<std::shared_ptr<EntryPoint>> entry_points;
 
@@ -652,14 +713,20 @@ struct Module {
     VulkanTypedHandle handle_;                            // Will be updated once its known its valid SPIR-V
     VulkanTypedHandle handle() const { return handle_; }  // matches normal convention to get handle
 
-    // Used for when modifying the SPIR-V (spirv-opt, GPU-AV instrumentation, etc) and need reparse it for VVL validaiton
-    Module(vvl::span<const uint32_t> code) : valid_spirv(true), words_(code.begin(), code.end()), static_data_(*this) {}
+    // Only currently used for when modifying the SPIR-V after spirv-opt and we need reparse it for VVL validation
+    explicit Module(vvl::span<const uint32_t> code)
+        : valid_spirv(true), words_(code.begin(), code.end()), static_data_(*this, true, nullptr) {}
 
+    // Used when we want to create a spirv::Module object (to make it easier to have a handle everywhere) but don't actually want to
+    // store/parse the SPIR-V itself (because it is turned off via settings)
+    explicit Module(bool is_valid_spirv) : valid_spirv(is_valid_spirv) {}
+
+    // "Normal" case
     // StatelessData is a pointer as we have cases were we don't need it and simpler to just null check the few cases that use it
-    Module(size_t codeSize, const uint32_t *pCode, StatelessData *stateless_data = nullptr)
-        : valid_spirv(pCode && pCode[0] == spv::MagicNumber && ((codeSize % 4) == 0)),
+    Module(size_t codeSize, const uint32_t *pCode, bool is_valid_spirv, bool parse, StatelessData *stateless_data)
+        : valid_spirv(is_valid_spirv),
           words_(pCode, pCode + codeSize / sizeof(uint32_t)),
-          static_data_(*this, stateless_data) {}
+          static_data_(*this, parse, stateless_data) {}
 
     const Instruction *FindDef(uint32_t id) const {
         auto it = static_data_.definitions.find(id);
@@ -681,27 +748,7 @@ struct Module {
         return (it != static_data_.execution_modes.end()) ? it->second : static_data_.empty_execution_mode;
     }
 
-    std::shared_ptr<const TypeStructInfo> GetTypeStructInfo(uint32_t struct_id) const {
-        // return the actual execution modes for this id, or a default empty set.
-        const auto it = static_data_.type_struct_map.find(struct_id);
-        return (it != static_data_.type_struct_map.end()) ? it->second : nullptr;
-    }
-    // Overload to walk down and find the OpTypeStruct
-    std::shared_ptr<const TypeStructInfo> GetTypeStructInfo(const Instruction *insn) const {
-        while (true) {
-            if (insn->Opcode() == spv::OpVariable) {
-                insn = FindDef(insn->Word(1));
-            } else if (insn->Opcode() == spv::OpTypePointer) {
-                insn = FindDef(insn->Word(3));
-            } else if (insn->IsArray()) {
-                insn = FindDef(insn->Word(2));
-            } else if (insn->Opcode() == spv::OpTypeStruct) {
-                return GetTypeStructInfo(insn->Word(1));
-            } else {
-                return nullptr;
-            }
-        }
-    }
+    std::shared_ptr<const TypeStructInfo> GetTypeStructInfo(const Instruction *insn) const;
 
     // Used to get human readable strings for error messages
     std::string GetDecorations(uint32_t id) const;
@@ -710,11 +757,10 @@ struct Module {
     void DescribeTypeInner(std::ostringstream &ss, uint32_t type, uint32_t indent) const;
     std::string DescribeType(uint32_t type) const;
     std::string DescribeVariable(uint32_t id) const;
+    std::string DescribeInstruction(const Instruction &error_insn) const;
 
-    std::optional<VkPrimitiveTopology> GetTopology(const EntryPoint &entrypoint) const;
-
-    std::shared_ptr<const EntryPoint> FindEntrypoint(char const *name, VkShaderStageFlagBits stageBits) const;
-    bool FindLocalSize(const EntryPoint &entrypoint, uint32_t &local_size_x, uint32_t &local_size_y, uint32_t &local_size_z) const;
+    std::shared_ptr<const EntryPoint> FindEntrypoint(const char *name, VkShaderStageFlagBits stageBits) const;
+    LocalSize FindLocalSize(const EntryPoint &entrypoint) const;
 
     uint32_t CalculateWorkgroupSharedMemory() const;
 
@@ -732,10 +778,14 @@ struct Module {
     uint32_t GetTypeBytesSize(const Instruction *insn) const;
     uint32_t GetBaseType(const Instruction *insn) const;
     const Instruction *GetBaseTypeInstruction(uint32_t type) const;
+    const Instruction *GetVariablePointerType(const spirv::Instruction &var_insn) const;
+    const Instruction *GetVariableDataType(const spirv::Instruction &var_insn) const;
     uint32_t GetTypeId(uint32_t id) const;
     uint32_t GetTexelComponentCount(const Instruction &insn) const;
     uint32_t GetFlattenArraySize(const Instruction &insn) const;
     AtomicInstructionInfo GetAtomicInfo(const Instruction &insn) const;
+    spv::StorageClass StorageClass(const Instruction &insn) const;
+    bool UsesStorageCapabilityStorageClass(const Instruction &insn) const;
 
     bool HasCapability(spv::Capability find_capability) const {
         return std::any_of(static_data_.capability_list.begin(), static_data_.capability_list.end(),
@@ -745,18 +795,23 @@ struct Module {
 
 }  // namespace spirv
 
+struct GlobalSettings;
+
 // Represents a VkShaderModule handle
 namespace vvl {
+
+// Need to allow a way to not waste time copying over to spirv::Module::words_ when we don't want to store the SPIR-V
+std::shared_ptr<spirv::Module> CreateSpirvModuleState(size_t codeSize, const uint32_t *pCode, const GlobalSettings &global_settings,
+                                                      spirv::StatelessData *stateless_data = nullptr);
+
 struct ShaderModule : public StateObject {
-    ShaderModule(VkShaderModule handle, std::shared_ptr<spirv::Module> &spirv_module, uint32_t unique_shader_id)
-        : StateObject(handle, kVulkanObjectTypeShaderModule), spirv(spirv_module), gpu_validation_shader_id(unique_shader_id) {
+    ShaderModule(VkShaderModule handle, std::shared_ptr<spirv::Module> &spirv_module)
+        : StateObject(handle, kVulkanObjectTypeShaderModule), spirv(spirv_module) {
         spirv->handle_ = handle_;
     }
 
     // For when we need to create a module with no SPIR-V backing it
-    ShaderModule(uint32_t unique_shader_id)
-        : StateObject(static_cast<VkShaderModule>(VK_NULL_HANDLE), kVulkanObjectTypeShaderModule),
-          gpu_validation_shader_id(unique_shader_id) {}
+    ShaderModule() : StateObject(static_cast<VkShaderModule>(VK_NULL_HANDLE), kVulkanObjectTypeShaderModule) {}
 
     VkShaderModule VkHandle() const { return handle_.Cast<VkShaderModule>(); }
 
@@ -765,7 +820,8 @@ struct ShaderModule : public StateObject {
     // https://github.com/KhronosGroup/Vulkan-ValidationLayers/pull/6265/files
     std::shared_ptr<spirv::Module> spirv;
 
-    // Used as way to match instrumented GPU-AV shader to a VkShaderModule handle
-    uint32_t gpu_validation_shader_id = 0;
+    // Used by GPU-AV to make sure instrumentation only occurs in a single thread at a time
+    // (Currently we don't need seem to need this for ShaderObjects)
+    std::mutex module_mutex_;
 };
 }  // namespace vvl
